@@ -1,5 +1,7 @@
 #include <libxml++/parsers/textreader.h>
 #include <libxml++/exceptions/internal_error.h>
+#include <libxml++/exceptions/parse_error.h>
+#include <libxml++/exceptions/validity_error.h>
 
 #include <libxml/xmlreader.h>
 
@@ -24,7 +26,7 @@ TextReader::TextReader(
     struct _xmlTextReader* cobj)
     : propertyreader(new PropertyReader(*this)), impl_( cobj )
 {
-	
+  setup_exceptions();
 }
 
 TextReader::TextReader(
@@ -35,7 +37,13 @@ TextReader::TextReader(
 	  impl_( xmlReaderForMemory ((const char*)data, size, uri.c_str(), NULL, 0) )
 {
   if( ! impl_ )
+  {
+    #ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
     throw internal_error("Cannot instantiate underlying libxml2 structure");
+    #endif //LIBXMLCPP_EXCEPTIONS_ENABLED
+  }
+
+  setup_exceptions();
 }
 
 TextReader::TextReader(
@@ -43,7 +51,13 @@ TextReader::TextReader(
   : propertyreader(new PropertyReader(*this)), impl_( xmlNewTextReaderFilename(URI.c_str()) )
 {
   if( ! impl_ )
+  {
+    #ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
     throw internal_error("Cannot instantiate underlying libxml2 structure");
+    #endif //LIBXMLCPP_EXCEPTIONS_ENABLED
+  }
+
+  setup_exceptions();
 }
 
 TextReader::~TextReader()
@@ -126,19 +140,19 @@ bool TextReader::is_empty_element() const
 Glib::ustring TextReader::get_local_name() const
 {
   return propertyreader->String(
-      xmlTextReaderLocalName(impl_));
+      xmlTextReaderLocalName(impl_), true);
 }
 
 Glib::ustring TextReader::get_name() const
 {
   return propertyreader->String(
-      xmlTextReaderName(impl_));
+      xmlTextReaderName(impl_), true);
 }
 
 Glib::ustring TextReader::get_namespace_uri() const
 {
   return propertyreader->String(
-      xmlTextReaderNamespaceUri(impl_));
+      xmlTextReaderNamespaceUri(impl_), true);
 }
 
 TextReader::xmlNodeType TextReader::get_node_type() const
@@ -152,7 +166,7 @@ TextReader::xmlNodeType TextReader::get_node_type() const
 Glib::ustring TextReader::get_prefix() const
 {
   return propertyreader->String(
-      xmlTextReaderPrefix(impl_));
+      xmlTextReaderPrefix(impl_), true);
 }
 
 char TextReader::get_quote_char() const
@@ -327,14 +341,49 @@ bool TextReader::is_valid() const
       xmlTextReaderIsValid(impl_));
 }
 
+void TextReader::setup_exceptions()
+{
+  xmlTextReaderErrorFunc func = NULL;
+  void* arg = NULL; 
 
+  // We respect any other error handlers already setup:
+  xmlTextReaderGetErrorHandler(impl_, &func, &arg);
+  if(!func)
+  {
+     func = (xmlTextReaderErrorFunc)&TextReader::on_libxml_error;
+     xmlTextReaderSetErrorHandler(impl_, func, this);
+  }
+}
 
+void TextReader::on_libxml_error(void* arg, const char* msg, int severity, void* locator)
+{
+  TextReader* ths = (TextReader*)arg;
+  ths->severity_ = severity;
+  ths->error_ = msg ? msg : "unknown parse error";
+}
+
+void TextReader::check_for_exceptions() const
+{
+  if( severity_ == 0 )
+    return;
+    
+  TextReader* ths = const_cast<TextReader*>(this);
+
+  int severity = severity_;
+  ths->severity_ = 0;
+
+  if( severity == XML_PARSER_SEVERITY_ERROR )
+    throw parse_error(error_);
+  else if( severity == XML_PARSER_SEVERITY_VALIDITY_ERROR )
+    throw validity_error(error_);
+}
 
 int TextReader::PropertyReader::Int(
     int value)
 {
   if(value == -1)
     owner_.check_for_exceptions();
+
   return value;
 }
 
@@ -382,9 +431,4 @@ Glib::ustring TextReader::PropertyReader::String(
   return (const char *)value;
 }
 
-void TextReader::check_for_exceptions() const
-{
-  //TODO: Shouldn't we do something here? murrayc.
-}
-
 } // namespace xmlpp
